#include "util.h"

#define kind dynamic,50


inline double BiasRankUpdate(const smat_t &R, mat_t &W, mat_t &H, vec_t &oldWt, vec_t &oldHt, const double alpha, double lambda, const int nowt)
{
	double g=0, h=0, y=0;
	int k = W.size();

	vec_t &Wt = W[nowt], &Ht = H[nowt];


	double hh= 0;
	for (long c=0 ; c<R.rows ; ++c)
		hh += Wt[c]*Wt[c];

	vector<double>  wa(R.cols);
	
//#pragma omp parallel for  schedule(kind)  
	for (int c=0 ; c<R.cols ; c++ ) {
		wa[c] = 0;
		for (long idx = R.col_ptr[c] ; idx<R.col_ptr[c+1] ; idx++ ) {
			wa[c] += Wt[R.row_idx[idx]]*R.rating[idx];
		}
	}

	vector<double> ww(k);

//#pragma omp parallel for 
	for (int i=0 ; i<k ; i++ ) {
		ww[i] = 0;
		vec_t &Wi = W[i];
		for ( int j=0 ; j<R.rows ; j++ )
			ww[i] += Wt[j]*Wi[j];
	}

	vector<double> wwh(R.cols);
//	double *wwh = MALLOC(double, R.cols);
//#pragma omp parallel for  
	
	for ( int i=0 ; i<k ; i++ )
		if (i!=nowt)
	{
		double wwnow = ww[i];
		vector<double> &Hnow = H[i];
		for ( int c=0 ; c<R.cols ; c++ ) {
			wwh[c] += wwnow*Hnow[c];
		}
	}

/*	for (int c=0 ; c<R.cols ; c++ ) {
		wwh[c] = 0;
		for (int i=0 ; i<k ; i++ )
			if (i != nowt)
			wwh[c] += ww[i]*H[i][c];
	}*/

	// update here
//#pragma omp parallel for 
	for (long c=0 ; c<R.cols ; c++){
		double h=0, y=0, g=0;
		if(R.col_ptr[c+1]==R.col_ptr[c]) continue;
		for(long idx=R.col_ptr[c]; idx < R.col_ptr[c+1]; ++idx) {
			// missing value estimation
			int tmp = R.row_idx[idx];
			double Wti = Wt[tmp];
//			printf("Wti: %lf\n", Wti);
			h += Wti*Wti;
			y += Wti*oldWt[tmp]; 
			g +=Wti*R.val[idx];
		}
//		printf("c:%d g:%lf, y:%lf, h:%lf\n", c, g, y, h);
		H[nowt][c] = ((1-alpha)*(g+y*oldHt[c]) +alpha*(wa[c]-wwh[c]) )/( (1-alpha)*h+alpha*hh+lambda );
	}
}

inline double RankOneUpdate(const smat_t &R, const int c, const vec_t &Wt, const double lambda, const double oldHtj, const vec_t &oldWt){
	double g=0, h=0, y=0;
	if(R.col_ptr[c+1]==R.col_ptr[c]) return 0;
	for(long idx=R.col_ptr[c]; idx < R.col_ptr[c+1]; ++idx) {
		double Wti = Wt[R.row_idx[idx]];
//		printf("Wti: %lf\n", Wti);
//		printf("nn %d %lf %d\n", idx,R.val[idx], R.row_idx[idx]);
		h += Wti*Wti;
		y += Wti*oldWt[R.row_idx[idx]]; 
		g +=Wti*R.val[idx];
//		printf("end\n");
	}
//	printf("c:%d g:%lf, y:%lf, h:%lf\n", c, g, y, h);
	return (g+y*oldHtj)/(h+lambda);
}


/*
inline double RankOneUpdate(const smat_t &R, const int c, const vec_t &Wt, const double lambda, const double oldHtj){
	double g=0, h=0;
	if(R.col_ptr[c+1]==R.col_ptr[c]) return 0;
	for(long idx=R.col_ptr[c]; idx < R.col_ptr[c+1]; ++idx) {
		double Wti = Wt[R.row_idx[idx]];
		h += Wti*Wti;
		g +=Wti*R.val[idx];
	}
	return (g+h*oldHtj)/(h+lambda);
}
*/
inline double UpdateRating(smat_t &R, const vec_t &Wt, const vec_t &Ht, const vec_t &oldWt, const vec_t &oldHt) {
	double loss=0;
//#pragma omp parallel for  schedule(kind) reduction(+:loss)
		for(int c =0; c < R.cols; ++c){
			double Htc = Ht[c], oldHtc = oldHt[c], loss_inner = 0;
			for(long idx=R.col_ptr[c]; idx < R.col_ptr[c+1]; ++idx){
				R.val[idx] -=  Wt[R.row_idx[idx]]*Htc-oldWt[R.row_idx[idx]]*oldHtc;
				loss_inner += R.val[idx]*R.val[idx];
			}
			loss += loss_inner;
		}
	return loss;	
}

void ccd_bias(smat_t &R, mat_t &W, mat_t &H, double lambda, double alpha, int maxiter=10, int inneriter =5){
	smat_t Rt;
	double reg=0,loss;
	srand(0);
	int k = W.size();
	double timebegin = clock();
	
//	omp_set_num_threads(8);
	R.copy_rating();

	Rt = R.transpose();
	
	for (int i=0 ; i<H.size() ; i++ )
		for (int j=0 ; j<R.cols ; j++ )
			H[i][j] = 0;
	// initial reg
	for (int t=0 ; t<k ; t++) for (long r=0 ; r<R.rows ; r++) reg += W[t][r]*W[t][r];
	for(int oiter = 1; oiter <= maxiter; ++oiter) {
		for(int tt=0; tt < k; ++tt) {
			int t = tt;
			if(oiter>1) t = rand()%k;
			
			vec_t &Wt = W[t], &Ht = H[t];
			vec_t oldWt(Wt.size(),0), oldHt(Ht.size(),0);
			if(oiter>1) {
				for(int c = 0; c < R.cols;++c) oldHt[c] = Ht[c];
				for(int c = 0; c < Rt.cols; ++c) oldWt[c] = Wt[c];
			}
			else
				for(int c = 0; c < Rt.cols; ++c) oldWt[c] = Wt[c];
			for(int iter = 1; iter <= inneriter; ++iter){
				// Update H[t]
				// compute sum_i W_ti^2
				//
				
//	printf("adsfasdf\n");
				BiasRankUpdate(R, W, H, oldWt, oldHt, alpha, lambda, t);
				BiasRankUpdate(Rt,H,W,oldHt,oldWt,alpha,lambda, t);
//#pragma omp parallel for  schedule(kind) shared(R,H,W)
//				for(long c = 0; c < R.cols; ++c)
//					Ht[c] = RankOneUpdate(R, c, Wt, lambda, oldHt[c],oldWt);
//				printf("after H update\n");
				// Update W[t]
//#pragma omp parallel for  schedule(kind) shared(R,H,W)
//				for(long c = 0; c < Rt.cols; ++c)
//					Wt[c] = RankOneUpdate(Rt, c, Ht, lambda, oldWt[c],oldHt);
				//	printf("iter %d-%d obj %.6g ", t, iter, calobj(R, W, H, lambda, true));
				//	puts("");
//				for (int c=0 ; c<R.cols ; c++)
//					printf("%lf ", H[0][c]);
//				printf("\n");

			}
			// Update R and Rt
			loss = UpdateRating(R, Wt, Ht, oldWt, oldHt);
			loss = UpdateRating(Rt, Ht, Wt, oldHt, oldWt);
			// function value for all loss
			double loss2 = 0;
			
/*	
 			#pragma omp parallel for
 			for (long c=0 ; c<R.cols ; c++ ){
				long now = 0;
				for (long idx = R.col_ptr[c] ; idx<R.col_ptr[c+1] ; idx++) {
					int ii = R.row_idx[idx];
					for (; now<ii ; now++)
					{
						double tmp = 0;
						for (int i=0 ; i<k ; i++)
							tmp += W[i][now]*H[i][c];
						loss2 += tmp*tmp;
					}
					now = ii+1;
				}
				for ( ; now<R.rows ; now++) {
					double tmp = 0;
					for (int i=0 ; i<k ; i++)
						tmp += W[i][now]*H[i][c];
					loss2 += tmp*tmp;
				}
			}*/
			for(long c = 0; c < R.cols; ++c) reg += (Ht[c]*Ht[c]-oldHt[c]*oldHt[c]);
			for(long c = 0; c < Rt.cols; ++c) reg += (Wt[c]*Wt[c]-oldWt[c]*oldWt[c]);
//			printf("oiter %d rank %d obj %.6g time %lf\n",oiter,t+1, loss + loss2*alpha+ reg*lambda, (clock()-timebegin)/CLOCKS_PER_SEC);
			fflush(stdout);
		}
		printf("Finished outer_iter %d\n", oiter);
	}
}


void usage(){ puts("export OMP_NUM_THREADS=4; ./omp-ccd-r1 rank lambda maxiter srcdir"); }
/*
int main(int argc, char* argv[]){
	if(argc != 5) { usage(); return -1;}
	int k = atoi(argv[1]); // rank
	double lambda = atof(argv[2]);
	int maxiter = atoi(argv[3]);
	char* src = argv[4];
	smat_t R;
	mat_t W,H;
	testset_t T;
	load(src,R,T);
	initial(W, k, R.rows);
	initial(H, k, R.cols);

	omp_set_num_threads(8);
	puts("starts!");
	double time = omp_get_wtime();
	ccd(R, W, H, lambda, T, maxiter);
	printf("Wall-time: %lg secs\n", omp_get_wtime() - time);
	return 0;
}
*/
