#include "mex.h"
#include <math.h>
#include <time.h>
#include "util.h"
#include "ccd-r1-cyclic.cpp"
//#include "als.cpp"

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
	int m, n;
    int i, j, k, maxiter;
	double rho, alpha;
	double *Wout, *Hout, *A;
	smat_t R;
    
	if (nlhs > 2 || nrhs != 5) {
		mexErrMsgTxt("Usage: [W,H] = matFact_mex(A, k, maxiter, rho, alpha) \n");
	}
	
	// input arguments
	m = mxGetM(prhs[0]);
	n = mxGetN(prhs[0]);
	
	mwIndex *ir, *jc;
	ir = mxGetIr(prhs[0]);
	jc = mxGetJc(prhs[0]);

	double *values;
	long nnz = mxGetNzmax(prhs[0]);
	values = mxGetPr(prhs[0]);

	k = (int)mxGetScalar(prhs[1]);
	maxiter = (int)mxGetScalar(prhs[2]);
	rho = mxGetScalar(prhs[3]);
	alpha = mxGetScalar(prhs[4]);

    //output arguments
    plhs[0] = mxCreateDoubleMatrix(k, m, mxREAL);
    Wout = mxGetPr(plhs[0]);
    
    plhs[1] = mxCreateDoubleMatrix(k, n, mxREAL);
    Hout = mxGetPr(plhs[1]);
    
	//create smat_t R
	R.rows = m; R.cols = n;
	R.nnz = nnz;

	R.val = MALLOC(double, nnz); R.val_t = MALLOC(double, nnz);
	R.row_idx = MALLOC(unsigned, nnz); R.col_idx = MALLOC(unsigned, nnz);
	R.row_ptr = MALLOC(long, m+1); R.col_ptr = MALLOC(long, n+1);
	

	// rows/columns start from 0
	for ( long i=0 ; i<nnz ; i++) {
		R.val[i] = values[i];
		R.row_idx[i] = ir[i];
	}
	for ( int i=0 ; i<=n ; i++)
		R.col_ptr[i] = jc[i];


	for ( int i=0 ; i<=m ; i++ )
		R.row_ptr[i] = 0;
	for (long i=0 ; i<nnz ; i++ )
		R.row_ptr[ir[i]]++;
	for ( int i=1 ; i<m ; i++ )
		R.row_ptr[i] = R.row_ptr[i]+R.row_ptr[i-1];
	for ( int i=m ; i>=1 ; i--)
		R.row_ptr[i] = R.row_ptr[i-1];
	R.row_ptr[0] = 0;

	for (int i=0 ; i<n ; i++ )
		for ( long j=jc[i] ; j<jc[i+1] ; j++ )
		{
			R.val_t[R.row_ptr[ir[j]]] = values[j];
			R.col_idx[R.row_ptr[ir[j]]] = i;
			R.row_ptr[ir[j]]++;
		}

	for ( int i=m ; i>=1 ; i--)
		R.row_ptr[i] = R.row_ptr[i-1];
	R.row_ptr[0] = 0;
/*
	for ( int i=0 ; i<n ; i++ )
		for (long j=R.row_ptr[i] ; j<R.row_ptr[i+1]; j++ )
			printf("%d %d %lf\n", i, R.col_idx[j], R.val_t[j]);
	printf("\n");
*/
	// initialize W and H
	mat_t W,H;
	W = mat_t(k, vec_t(m));
	srand48(0);
	for (long i=0 ; i<k ; i++ )
		for (long j=0 ; j<m ; j++ )
			W[i][j] = 0.1*drand48();
	H = mat_t(k,vec_t(n));
	for (long i=0 ; i<k ; i++ )
		for (long j=0 ; j<n ; j++ )
			H[i][j] = 0.1*drand48();
//	initial(W, k, R.rows);
//	initial(W,(long)k,(long)m);
//	initial(H,k,n);

//	printf("maxiter: %d\n", maxiter);

//	printf("begin ccd_bias\n");
	ccd_bias(R,W,H,rho, alpha, maxiter);
//	printf("end ccd_bias\n");
	
	// set wout and hout
	for (int i=0, r=0 ; i<m ; i++)
		for (int j=0 ; j<k ; j++ )
			Wout[r++] = W[j][i];
	for (int i=0, r=0 ; i<n ; i++)
		for (int j=0 ; j<k ; j++ )
			Hout[r++] = H[j][i];
}


   
